==> slices_cli-0.1.81rc24-py3-none-any.whl: slices_cli/experiment_member.py <==
"""Experiment CLI."""

from enum import Enum
from typing import Annotated

import typer
import yaml
from pydantic import RootModel
from rich.table import Table
from slices_expauth_client.exceptions import ApiException
from slices_expauth_client.models.experiment_member_create import ExperimentMemberCreate
from slices_expauth_client.models.experiment_member_role import ExperimentMemberRole
from slices_expauth_client.models.experiment_member_update import ExperimentMemberUpdate

from . import console, err_console
from .context import Context
from .users_info import USER_ID_HELP, EnrichedExperimentMember
from .utils import get_error_detail, is_experiment_id

experiment_member_app = typer.Typer(
    help="""Commands to list/set/remove experiment members.

For some commands, you need the ID of other users.
You can find these with the command: slices project show""",
    short_help="Commands to list/set/remove experiment members",
    no_args_is_help=True,
)


def validate_user_id(ctx: typer.Context, value: str):
    """Validate user ID, resolving it if it is not a TypeID. Always return TypeID."""
    return ctx.obj.other_users_info.resolve_project_member(ctx.obj.require_project_id(), value)


def _get_experiment_id(ctx: Context, name: str) -> str:
    if is_experiment_id(name):
        return name
    experiments = ctx.obj.api.experiments_api.get_experiments(
        project_id=ctx.obj.require_project_id(),
        friendly_name=name,
    )

    if not len(experiments.experiments) > 0:
        err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
        raise typer.Exit(1)

    return experiments.experiments[0].id


ALL_ROLES = [k.lower() for k in ExperimentMemberRole.__members__]
ALL_ALLOWED_ROLES = [k for k in ALL_ROLES if k != "lead"]
ALL_ALLOWED_ROLES_STR = "/".join(ALL_ALLOWED_ROLES)


def validate_role(value: str) -> str:
    """Validate the experiment member role."""

    value = value.lower()
    if value not in ALL_ROLES:
        raise typer.BadParameter(f"Invalid role {value!r}. Allowed roles: {'/'.join(ALL_ALLOWED_ROLES)}")
    if value not in ALL_ALLOWED_ROLES:
        raise typer.BadParameter(
            f"May not change role of user in experiment to {value!r}. Allowed roles: {'/'.join(ALL_ALLOWED_ROLES)}"
        )
    return value


def _map_project_role_to_experiment_role(project_role: str) -> ExperimentMemberRole:
    """Explicit mapping of project roles to experiment roles."""

    # project has member roles: 'lead', 'admin', 'member', 'viewer'
    if project_role.lower() == "lead":
        return ExperimentMemberRole.ADMIN
    if project_role.lower() == "admin":
        return ExperimentMemberRole.ADMIN
    if project_role.lower() == "member":
        return ExperimentMemberRole.MEMBER
    if project_role.lower() == "viewer":
        return ExperimentMemberRole.VIEWER
    raise ValueError(f"Unsupported project role {project_role!r}")


def do_add_missing_project_members_to_experiment(
    ctx: Context, _experiment_name: str, experiment_id: str, *, print_progress: bool = False
):
    """Add all project members that are not in the experiment yet.

    Their role is based on their role in the project.
    """
    current_project_id = ctx.obj.get_current_project_id()

    if not current_project_id:
        err_console.print("[red]No project set. :x:")
        raise typer.Exit

    project_members_response = ctx.obj.api.projects_api.get_project_members(
        current_project_id,
    )
    project_members = project_members_response.project_members

    has_err = False
    for member in project_members:
        if ctx.obj.user_info and member.user_id == ctx.obj.user_info["sub"]:
            # Skip the user itself
            continue
        try:
            _experiment = ctx.obj.api.experiments_api.add_experiment_member(
                experiment_id,
                ExperimentMemberCreate(role=_map_project_role_to_experiment_role(member.role), user_id=member.user_id),
            )
            if print_progress and console.is_terminal:
                console.print(
                    f"[green]:heavy_check_mark: "
                    f"[bold]{ctx.obj.other_users_info.try_get_user_full_name(member.user_id)}[/bold] added "
                    f"with role [italic]{member.role.upper()}[/italic]",
                )
        except ApiException as err:
            if err.status == 409:  # noqa PLR2004: HTTP_409_CONFLICT
                if print_progress and console.is_terminal:
                    console.print(
                        f":heavy_check_mark: "
                        f"[bold]{ctx.obj.other_users_info.try_get_user_full_name(member.user_id)}[/bold] "
                        f"was already a member"
                    )
            else:
                err_console.print(f"[bold red]Could not add experiment member {member.user_id} :x:")
                err_console.print(f"[red]{get_error_detail(err.body)}")
                has_err = True

    if has_err:
        raise typer.Exit(1)


@experiment_member_app.command(name="set", rich_help_panel="Commands")
def set_experiment_member(
    ctx: Context,
    experiment_name: Annotated[
        str,
        typer.Argument(
            help="Name or ID of the experiment.",
            show_default=False,
        ),
    ],
    user_id: Annotated[
        str,
        typer.Argument(
            help=f"ID of the user. {USER_ID_HELP}",
            callback=validate_user_id,
            show_default=False,
        ),
    ],
    role: Annotated[
        str,
        typer.Argument(help=f"Role of the user ({ALL_ALLOWED_ROLES_STR})", callback=validate_role),
    ] = "member",
):
    """Add or update an experiment member.

    You can find the user ID of other users with the command: slices project show
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        added_experiment_member = ctx.obj.api.experiments_api.add_experiment_member(
            experiment_id,
            ExperimentMemberCreate(role=ExperimentMemberRole(role), user_id=user_id),
        )
        if console.is_terminal:
            console.print(
                f"[green]:heavy_check_mark: [bold]{ctx.obj.other_users_info.try_get_user_full_name(user_id)}[/bold] "
                f"added with role [italic]{added_experiment_member.role.name}[/italic] ",
            )
    except ApiException as err:
        if err.status == 409:  # noqa PLR2004: HTTP_409_CONFLICT
            try:
                updated_experiment_member = ctx.obj.api.experiments_api.update_experiment_member(
                    user_id,
                    experiment_id,
                    ExperimentMemberUpdate(role=ExperimentMemberRole(role)),
                )
                if console.is_terminal:
                    console.print(
                        f"[green]:heavy_check_mark: "
                        f"[bold]{ctx.obj.other_users_info.try_get_user_full_name(user_id)}[/bold] "
                        f"updated to role [italic]{updated_experiment_member.role.name}[/italic] ",
                    )
            except ApiException as err2:
                err_console.print("[bold red]Could not update experiment member :x:")
                err_console.print(f"[red]{get_error_detail(err2.body)}")
                raise typer.Exit(1) from err2
        else:
            err_console.print("[bold red]Could not add experiment member :x:")
            err_console.print(f"[red]{get_error_detail(err.body)}")
            raise typer.Exit(1) from err


@experiment_member_app.command(name="add-project-members", rich_help_panel="Commands")
def add_project_members_to_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
):
    """Add project members to the experiment (if not yet added).

    This does not change the role of existing project members in the experiment.
    The role in the experiment is based on the role in the project.
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)
    do_add_missing_project_members_to_experiment(ctx, experiment_name, experiment_id, print_progress=True)


@experiment_member_app.command(name="transfer-lead", rich_help_panel="Advanced")
def transfer_experiment_lead(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    user_id: Annotated[
        str,
        typer.Argument(
            help=f"ID of the new LEAD user. {USER_ID_HELP}",
            callback=validate_user_id,
        ),
    ],
):
    """Transfer the LEAD of the experiment from you to another member of the experiment.

    You can find the user IDs with the command: slices experiment member list
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        _experiment = ctx.obj.api.experiments_api.update_experiment_member(
            user_id,
            experiment_id,
            ExperimentMemberUpdate(role=ExperimentMemberRole(ExperimentMemberRole.LEAD)),
        )
        if console.is_terminal:
            console.print(
                f"[green]:heavy_check_mark: Transferred experiment {experiment_name!r} LEAD to [bold]{user_id}[/bold] "
            )
    except ApiException as err:
        err_console.print("[bold red]Could not transfer experiment LEAD (are you the LEAD?) :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err


class ListFormat(Enum):
    """List format options."""

    text = "text"
    json = "json"
    yaml = "yaml"


ExperimentsMembers = RootModel[list[EnrichedExperimentMember]]


@experiment_member_app.command(name="list", rich_help_panel="Commands")
def list_experiment_members(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    show_ids: Annotated[bool, typer.Option("--show-ids/--no-show-ids", "-I", help="Show internal user IDs.")] = True,
    format: Annotated[ListFormat, typer.Option("--format", "-f")] = "text",  # type: ignore
):
    """List experiment members."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    all_experiment_members_extra = ctx.obj.other_users_info.get_experiment_members(experiment_id)

    if format == ListFormat.json:
        console.print_json(
            ExperimentsMembers(all_experiment_members_extra).model_dump_json(by_alias=True, exclude_unset=True)
        )
        return
    if format == ListFormat.yaml:
        console.print(
            yaml.safe_dump(
                ExperimentsMembers(all_experiment_members_extra).model_dump(
                    mode="json", by_alias=True, exclude_unset=True
                )
            )
        )
        return

    if show_ids:
        table = Table("User ID", "Name", "Role", title=f"Members of experiment {experiment_name}")
        for member in all_experiment_members_extra:
            table.add_row(member.user_id, member.full_name, member.role.upper())
    else:
        table = Table("User", "Role", title=f"Members of experiment {experiment_name}")
        for member in all_experiment_members_extra:
            table.add_row(member.full_name, member.role.upper())

    console.print(table)


@experiment_member_app.command("remove", rich_help_panel="Commands")
def remove_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(..., help="Name or ID of experiment.")],
    user_id: Annotated[str, typer.Argument(..., help="User ID to remove from experiment.", callback=validate_user_id)],
):
    """Remove an experiment member."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        deleted_experiment_member = ctx.obj.api.experiments_api.delete_experiment_member(user_id, experiment_id)
    except ApiException as err:
        err_console.print("[bold red]Could not remove experiment member:x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Removed experiment member "
        f"'{ctx.obj.other_users_info.try_get_user_full_name(deleted_experiment_member.user_id)}' "
        f"from experiment {experiment_name}."
    )

==> slices_cli-0.1.82rc1-py3-none-any.whl: slices_cli/experiment_member.py <==
"""Experiment CLI."""

from enum import Enum
from typing import Annotated

import typer
import yaml
from pydantic import RootModel
from rich.table import Table
from slices_expauth_client.exceptions import ApiException
from slices_expauth_client.models.experiment_member_create import ExperimentMemberCreate
from slices_expauth_client.models.experiment_member_role import ExperimentMemberRole
from slices_expauth_client.models.experiment_member_update import ExperimentMemberUpdate

from . import console, err_console
from .context import Context
from .users_info import USER_ID_HELP, EnrichedExperimentMember
from .utils import get_error_detail, is_experiment_id

experiment_member_app = typer.Typer(
    help="""Commands to list/set/remove experiment members.

For some commands, you need the ID of other users.
You can find these with the command: slices project show""",
    short_help="Commands to list/set/remove experiment members",
    no_args_is_help=True,
)


def validate_user_id(ctx: typer.Context, value: str):
    """Validate user ID, resolving it if it is not a TypeID. Always return TypeID."""
    return ctx.obj.other_users_info.resolve_project_member(ctx.obj.require_project_id(), value)


def _get_experiment_id(ctx: Context, name: str) -> str:
    if is_experiment_id(name):
        return name
    experiments = ctx.obj.api.experiments_api.get_experiments(
        project_id=ctx.obj.require_project_id(),
        friendly_name=name,
    )

    if not len(experiments.experiments) > 0:
        err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
        raise typer.Exit(1)

    return experiments.experiments[0].id


ALL_ROLES = [k.lower() for k in ExperimentMemberRole.__members__]
ALL_ALLOWED_ROLES = [k for k in ALL_ROLES if k != "lead"]
ALL_ALLOWED_ROLES_STR = "/".join(ALL_ALLOWED_ROLES)


def validate_role(value: str) -> str:
    """Validate the experiment member role."""

    value = value.lower()
    if value not in ALL_ROLES:
        raise typer.BadParameter(f"Invalid role {value!r}. Allowed roles: {'/'.join(ALL_ALLOWED_ROLES)}")
    if value not in ALL_ALLOWED_ROLES:
        raise typer.BadParameter(
            f"May not change role of user in experiment to {value!r}. Allowed roles: {'/'.join(ALL_ALLOWED_ROLES)}"
        )
    return value


def _map_project_role_to_experiment_role(project_role: str) -> ExperimentMemberRole:
    """Explicit mapping of project roles to experiment roles."""

    # project has member roles: 'lead', 'admin', 'member', 'viewer'
    if project_role.lower() == "lead":
        return ExperimentMemberRole.ADMIN
    if project_role.lower() == "admin":
        return ExperimentMemberRole.ADMIN
    if project_role.lower() == "member":
        return ExperimentMemberRole.MEMBER
    if project_role.lower() == "viewer":
        return ExperimentMemberRole.VIEWER
    raise ValueError(f"Unsupported project role {project_role!r}")


def do_add_missing_project_members_to_experiment(
    ctx: Context, _experiment_name: str, experiment_id: str, *, print_progress: bool = False
):
    """Add all project members that are not in the experiment yet.

    Their role is based on their role in the project.
    """
    current_project_id = ctx.obj.get_current_project_id()

    if not current_project_id:
        err_console.print("[red]No project set. :x:")
        raise typer.Exit

    project_members_response = ctx.obj.api.projects_api.get_project_members(
        current_project_id,
    )
    project_members = project_members_response.project_members

    has_err = False
    for member in project_members:
        if ctx.obj.user_info and member.user_id == ctx.obj.user_info["sub"]:
            # Skip the user itself
            continue
        try:
            _experiment = ctx.obj.api.experiments_api.add_experiment_member(
                experiment_id,
                ExperimentMemberCreate(role=_map_project_role_to_experiment_role(member.role), user_id=member.user_id),
            )
            if print_progress and console.is_terminal:
                console.print(
                    f"[green]:heavy_check_mark: "
                    f"[bold]{ctx.obj.other_users_info.try_get_user_full_name(member.user_id)}[/bold] added "
                    f"with role [italic]{member.role.upper()}[/italic]",
                )
        except ApiException as err:
            if err.status == 409:  # noqa PLR2004: HTTP_409_CONFLICT
                if print_progress and console.is_terminal:
                    console.print(
                        f":heavy_check_mark: "
                        f"[bold]{ctx.obj.other_users_info.try_get_user_full_name(member.user_id)}[/bold] "
                        f"was already a member"
                    )
            else:
                err_console.print(f"[bold red]Could not add experiment member {member.user_id} :x:")
                err_console.print(f"[red]{get_error_detail(err.body)}")
                has_err = True

    if has_err:
        raise typer.Exit(1)


@experiment_member_app.command(name="set", rich_help_panel="Commands")
def set_experiment_member(
    ctx: Context,
    experiment_name: Annotated[
        str,
        typer.Argument(
            help="Name or ID of the experiment.",
            show_default=False,
        ),
    ],
    user_id: Annotated[
        str,
        typer.Argument(
            help=f"ID of the user. {USER_ID_HELP}",
            callback=validate_user_id,
            show_default=False,
        ),
    ],
    role: Annotated[
        str,
        typer.Argument(help=f"Role of the user ({ALL_ALLOWED_ROLES_STR})", callback=validate_role),
    ] = "member",
):
    """Add or update an experiment member.

    You can find the user ID of other users with the command: slices project show
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        added_experiment_member = ctx.obj.api.experiments_api.add_experiment_member(
            experiment_id,
            ExperimentMemberCreate(role=ExperimentMemberRole(role), user_id=user_id),
        )
        if console.is_terminal:
            console.print(
                f"[green]:heavy_check_mark: [bold]{ctx.obj.other_users_info.try_get_user_full_name(user_id)}[/bold] "
                f"added with role [italic]{added_experiment_member.role.name}[/italic] ",
            )
    except ApiException as err:
        if err.status == 409:  # noqa PLR2004: HTTP_409_CONFLICT
            try:
                updated_experiment_member = ctx.obj.api.experiments_api.update_experiment_member(
                    user_id,
                    experiment_id,
                    ExperimentMemberUpdate(role=ExperimentMemberRole(role)),
                )
                if console.is_terminal:
                    console.print(
                        f"[green]:heavy_check_mark: "
                        f"[bold]{ctx.obj.other_users_info.try_get_user_full_name(user_id)}[/bold] "
                        f"updated to role [italic]{updated_experiment_member.role.name}[/italic] ",
                    )
            except ApiException as err2:
                err_console.print("[bold red]Could not update experiment member :x:")
                err_console.print(f"[red]{get_error_detail(err2.body)}")
                raise typer.Exit(1) from err2
        else:
            err_console.print("[bold red]Could not add experiment member :x:")
            err_console.print(f"[red]{get_error_detail(err.body)}")
            raise typer.Exit(1) from err


@experiment_member_app.command(name="add-project-members", rich_help_panel="Commands")
def add_project_members_to_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
):
    """Add project members to the experiment (if not yet added).

    This does not change the role of existing project members in the experiment.
    The role in the experiment is based on the role in the project.
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)
    do_add_missing_project_members_to_experiment(ctx, experiment_name, experiment_id, print_progress=True)


@experiment_member_app.command(name="transfer-lead", rich_help_panel="Advanced")
def transfer_experiment_lead(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    user_id: Annotated[
        str,
        typer.Argument(
            help=f"ID of the new LEAD user. {USER_ID_HELP}",
            callback=validate_user_id,
        ),
    ],
):
    """Transfer the LEAD of the experiment from you to another member of the experiment.

    You can find the user IDs with the command: slices experiment member list
    """

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        _experiment = ctx.obj.api.experiments_api.update_experiment_member(
            user_id,
            experiment_id,
            ExperimentMemberUpdate(role=ExperimentMemberRole(ExperimentMemberRole.LEAD)),
        )
        if console.is_terminal:
            console.print(
                f"[green]:heavy_check_mark: Transferred experiment {experiment_name!r} LEAD to [bold]{user_id}[/bold] "
            )
    except ApiException as err:
        err_console.print("[bold red]Could not transfer experiment LEAD (are you the LEAD?) :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err


class ListFormat(Enum):
    """List format options."""

    text = "text"
    json = "json"
    yaml = "yaml"


ExperimentsMembers = RootModel[list[EnrichedExperimentMember]]


@experiment_member_app.command(name="list", rich_help_panel="Commands")
def list_experiment_members(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(help="Name or ID of the experiment.")],
    show_details: Annotated[
        bool,
        typer.Option(
            "--show-details/--no-show-details",
            "-I",
            help="Show internal user IDs, email, and usernames",
        ),
    ] = True,
    format: Annotated[ListFormat, typer.Option("--format", "-f")] = "text",  # type: ignore
):
    """List experiment members."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    all_experiment_members_extra = ctx.obj.other_users_info.get_experiment_members(experiment_id)

    if format == ListFormat.json:
        console.print_json(
            ExperimentsMembers(all_experiment_members_extra).model_dump_json(by_alias=True, exclude_unset=True)
        )
        return
    if format == ListFormat.yaml:
        console.print(
            yaml.safe_dump(
                ExperimentsMembers(all_experiment_members_extra).model_dump(
                    mode="json", by_alias=True, exclude_unset=True
                )
            )
        )
        return

    if show_details:
        table = Table("User ID", "Name", "Username", "Email", "Role", title=f"Members of experiment {experiment_name}")
        for member in all_experiment_members_extra:
            table.add_row(
                member.user_id, member.full_name, member.preferred_username, member.email, member.role.upper()
            )
    else:
        table = Table("User", "Role", title=f"Members of experiment {experiment_name}")
        for member in all_experiment_members_extra:
            table.add_row(member.full_name, member.role.upper())

    console.print(table)


@experiment_member_app.command("remove", rich_help_panel="Commands")
def remove_experiment_member(
    ctx: Context,
    experiment_name: Annotated[str, typer.Argument(..., help="Name or ID of experiment.")],
    user_id: Annotated[str, typer.Argument(..., help="User ID to remove from experiment.", callback=validate_user_id)],
):
    """Remove an experiment member."""

    experiment_id = _get_experiment_id(ctx, experiment_name)

    try:
        deleted_experiment_member = ctx.obj.api.experiments_api.delete_experiment_member(user_id, experiment_id)
    except ApiException as err:
        err_console.print("[bold red]Could not remove experiment member:x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Removed experiment member "
        f"'{ctx.obj.other_users_info.try_get_user_full_name(deleted_experiment_member.user_id)}' "
        f"from experiment {experiment_name}."
    )
